# This is a song I composed! It's very complex here in the file because it's a very complex tune.
# The song is produced by the function main_tune.

# pylint: disable=undefined-variable,wildcard-import,unused-wildcard-import,invalid-slice-index
import sys
import sound
from sound.notes import *
from sound.instrument import Instrument, Guitar
from sound.tone import Digitar, SquareWave, SAMPLE_RATE
from sound.signal import SliceSignal, Purifier, ReverseSignal, ConstantSignal

def strum_stuff(guitar, chord, length, bpm, basebeat=0):
    base = basebeat * SAMPLE_RATE * 60 / bpm
    L1 = length * 0.25 * SAMPLE_RATE * 60 / bpm
    L2 = length * 0.125 * SAMPLE_RATE * 60 / bpm
    guitar.queue(base, (guitar.strum_down, (chord,)))
    guitar.queue(base + L1, (guitar.strum_up, (chord,)))
    guitar.queue(base + L1 + L2, (guitar.strum_down, (chord,)))
    guitar.queue(base + L1 + L2 + L1, (guitar.strum_up, (chord,)))
    guitar.queue(base + L1 + L2 + L1 + L2, (guitar.strum_down, (chord,)))
    guitar.queue(base + L1 + L2 + L1 + L2 + L2, (guitar.strum_up, (chord,)))

def n_of_them(them, n):
    o = them
    for _ in range(n - 1):
        o &= them
    return o

def sampslice(src, start, length):
    return SliceSignal(src, float(start)/SAMPLE_RATE, float(length)/SAMPLE_RATE, True)

def pitch_right(src, shift):
    ps = sound.filter.PitchShift(src, shift + 1)
    dat = [None]*int(src.duration*10)
    i = 0
    while not ps.done:
        dat[i] = ps.amplitude(i)
        i += 1
    ob = Purifier(0, 0)
    ob.storage = dat[:i]
    ob.duration = i
    ob.src = src
    ob.nextf = i
    return ob

def glitchmobile(src, *args):
    cut = False
    src = Purifier(src)
    pieces = []
    current = 0
    for arg in args:
        if arg[0] == 'ok':
            _, length = arg
            length = int(length*SAMPLE_RATE)
            pieces.append(sampslice(src, current, length))
            current += length
        elif arg[0] == 'loop':
            _, length, times = arg
            length = length * SAMPLE_RATE
            piece = sampslice(src, current, length)
            for _ in range(times):
                pieces.append(piece)
            current += length
        elif arg[0] == 'skip':
            _, length = arg
            current += length * SAMPLE_RATE
        elif arg[0] == 'gap':
            _, length = arg
            piece = ConstantSignal(0)
            piece.duration = length * SAMPLE_RATE
            pieces.append(piece)
        elif arg[0] == 'from':
            _, length, where = arg
            length *= SAMPLE_RATE
            where = where*SAMPLE_RATE + current
            piece = sampslice(src, where, length)
            pieces.append(piece)
        elif arg[0] == 'pitch':
            _, length, pitch = arg
            length *= SAMPLE_RATE
            piece = sampslice(src, current, length)
            pieces.append(pitch_right(piece, pitch))
            current += piece.duration
        elif arg[0] == 'pitchfrom':
            _, length, pitch, where = arg
            length *= SAMPLE_RATE
            where = where*SAMPLE_RATE + current
            piece = sampslice(src, where, length)
            pieces.append(pitch_right(piece, pitch))
        elif arg[0] == 'reversefrom':
            _, length, where = arg
            length *= SAMPLE_RATE
            where = where*SAMPLE_RATE + current
            piece = sampslice(src, where, length)
            pieces.append(ReverseSignal(piece))
        elif arg[0] == 'cut':
            cut = True
            break

    if current < src.duration and not cut:
        pieces.append(sampslice(src, current, src.duration - current))
    o = pieces[0]
    for p in pieces[1:]:
        o &= p
    return o

class Guitar2(Instrument):
    def __init__(self, sample):
        self.sample = sample
        super(Guitar2, self).__init__()

    def _note(self, freq, beats, filling, legato):
        digibase = self.sample(freq)
        return digibase * sound.envelope.Envelope(beats * self.beat * filling)

def make_digi(f, bufs, length):
    return Digitar(f, buffersize=bufs)[:length]

def main_tune():
    guitar = sound.asyncplayer.GuitarStrummer(Digitar)
    strum_stuff(guitar, 'C', 2, 140)
    strum_stuff(guitar, 'G', 2, 140, 2)
    strum_stuff(guitar, 'Am', 2, 140, 4)
    strum_stuff(guitar, 'F', 2, 140, 6)
    strum_stuff(guitar, 'C', 2, 140, 8)
    strum_stuff(guitar, 'G', 2, 140, 10)
    strum_stuff(guitar, 'Am', 2, 140, 12)
    strum_stuff(guitar, 'F', 2, 140, 14)

    buf = guitar[: 60./140*8].purify()
    buf2 = guitar[60./140*8 : 60./140*16].purify()
    p1 = buf[: 60./140*6]
    p2 = buf[60./140*6 : 60./140*6 + 0.1]
    p3 = p2[:0.2]
    p4 = p1 + (p3 >> 60./140*4.4)*2
    p5 = n_of_them(p3, 21)
    p6 = sound.tone.BrownNoise() * sound.envelope.Line(0, 3, 3)
    p7 = p5 + (p6 >> 1.2)
    p8 = (p4 & p7) / 8

    gi = Guitar()
    gi.tempo = 140
    gn = gi.note
    m1 = gn(C4) & gn(G3, 0.75) & gn(G3, 0.25) & gn(A3, 0.5) & gn(G3, 0.5) & gn(F3, 0.5) & gn(E3, 0.5) & gn(E3, 0.25) & gn(F3, 0.25) & gn(G3, 1) & gn(C3, 0.5) & gn(D3, 1.5) & gn(C3, 0.5)
    m2 = gn(C4) & gn(G3, 0.75) & gn(G3, 0.25) & gn(A3, 0.5) & gn(G3, 0.5) & gn(F3, 0.5) & gn(E3, 0.5) & gn(C4) & gn(B3) & gn(C4, 2)

    m3 = ((buf&buf2) / 8 + (m1&m2) / 2)
    m4 = glitchmobile(m3, ('ok', 2.2), ('loop', 0.01, 40), ('gap', 0.1), ('ok', 0.9), ('gap', 0.1), ('ok', 0.5), ('loop', 0.5, 3), ('reversefrom', 1, 0), ('ok', 1.1), ('gap', 0.3), ('ok', 0.5), ('pitch', 0.5, sound.envelope.Line(0, 3, 1)))

    m5 = gn(C4) & gn(E4, 0.75) & gn(C4, 0.25) & gn(F4, 0.5) & gn(E4, 0.5) & gn(D4, 0.5) & gn(C4, 0.5) & gn(D4, 0.25) & gn(E4, 0.25) & gn(D4, 0.25) & gn(E4, 0.25)  & gn(D4, 0.25) & gn(E4, 0.25) & gn(D4, 0.25) & gn(E4, 0.25) & gn(D4, 1)
    bm5 = Purifier(buf2/8 + m5 / 2)
    line = sound.envelope.Line(0, 5, 60./140*8)
    square = SquareWave(50)*sound.envelope.Envelope(0.5)
    m6 = pitch_right(bm5, sound.tone.Noise() * line * line * line) & square

    md = make_digi
    n1 = md(C4 - 10, 50, 1)
    for i in range(30):
        n1 &= md(C4 - 10 + i/2., 50 + i*19, (2**-(i/7.)))

    guitar = sound.asyncplayer.GuitarStrummer(lambda f: Digitar(f, buffersize=350))
    guitar.strum_down('C', 23000)
    guitar.queue(23000*6.5, (guitar.strum_down, ('F',400)))
    buf = guitar[:float(n1.duration)/SAMPLE_RATE - 6].purify(True)
    n2 = (buf/3 >> 6) + n1 / 2
    n3 = n2 >> 1.5

    startf = guitar.frame
    strum_stuff(guitar, 'C', 2, 140, 0)
    strum_stuff(guitar, 'G', 2, 140, 2)
    strum_stuff(guitar, 'Am', 2, 140, 4)
    strum_stuff(guitar, 'F', 2, 140, 6)
    buf2 = SliceSignal(guitar, float(startf)/SAMPLE_RATE, 60./140*8, True).purify(True)

    gi = Guitar2(lambda f: Digitar(f, buffersize=350))
    gi.tempo = 140
    gn = gi.note
    q1 = gn(C4) & gn(G3, 0.75) & gn(G3, 0.25) & gn(A3, 0.5) & gn(G3, 0.5) & gn(F3, 0.5) & gn(E3, 0.5) & gn(E3, 0.25) & gn(F3, 0.25) & gn(G3, 1) & gn(C3, 0.5) & gn(D3, 1.5) & gn(C3, 0.5)
    q2 = gn(C4) & gn(G3, 0.75) & gn(G3, 0.25) & gn(A3, 0.5) & gn(G3, 0.5) & gn(F3, 0.5) & gn(E3, 0.5) & gn(C4) & gn(B3) & gn(C4, 2)
    q3 = ((buf2&buf2) / 6 + (q1&q2) / 2)
    q4 = glitchmobile(q3, ('ok', 0.5), ('loop', 0.1, 3), ('ok', 0.5), ('loop', 0.1, 5), ('ok', 0.5), ('loop', 0.2, 4), ('ok', 0.3), ('pitch', 0.3, -0.5), ('pitch', 0.5, -0.2), ('ok', 0.5), ('gap', 0.1), ('ok', 0.1), ('gap', 0.1), ('pitchfrom', 3, 2, 0), ('ok', 1), ('loop', 0.1, 7), ('pitch', 0.5, -0.2), ('pitch', 0.5, -0.4), ('pitch', 0.25, -0.6), ('pitch', 0.25, -0.8), ('cut',))
    q4_5 = Purifier(Digitar(C4, buffersize=350), 6)
    q5 = sound.filter.PitchShift(q4_5, 0.1) * sound.envelope.ADSR(0, 0, 4, 2, sustain_level=1)
    q6 = q4 & q5

    e1 = pitch_right(Purifier(gi.note(C3, 5)), sound.tone.Noise()*1) >> 1
    e2 = pitch_right(Purifier(gi.note(D3, 10)), sound.tone.Noise()*2)
    e3 = q5[:3.5]
    e4 = (e1 & e2) + (e3/2 >> 4)
    e5 = e4 >> 0.5

    f1 = pitch_right(Purifier(Digitar(C3, buffersize=400), 1), sound.tone.Noise()*2)
    f2 = pitch_right(Purifier(Digitar(C3, buffersize=600), 1), sound.tone.Noise()*1.2)
    f3 = pitch_right(Purifier(Digitar(C3, buffersize=800), 0.8), sound.tone.Noise()*1)
    f4 = pitch_right(Purifier(Digitar(C3, buffersize=1200), 0.5), sound.tone.Noise()*0.5)
    f5 = pitch_right(Purifier(Digitar(C3, buffersize=1400), 0.5), sound.tone.Noise()*0.3)
    f6 = pitch_right(Purifier(Digitar(C3, buffersize=1600), 0.5), sound.tone.Noise()*0.1)
    f7 = Purifier(Digitar(C3, buffersize=2000), 2)
    f8 = f1 & f2 & f3 & f4 & f5 & f6 & f7

    guitar = sound.asyncplayer.GuitarStrummer(lambda f: sound.tone.SawtoothWave(f) * sound.envelope.envelope(decay=0.01))
    strum_stuff(guitar, 'C', 2, 160)
    strum_stuff(guitar, 'G', 2, 160, 2)
    strum_stuff(guitar, 'Am', 2, 160, 4)
    strum_stuff(guitar, 'F', 2, 160, 6)
    strum_stuff(guitar, 'C', 2, 160, 8)
    strum_stuff(guitar, 'G', 2, 160, 10)
    strum_stuff(guitar, 'Am', 2, 160, 12)
    strum_stuff(guitar, 'F', 2, 160, 14)
    strum_stuff(guitar, 'C', 2, 160, 16)
    strum_stuff(guitar, 'G', 2, 160, 18)
    strum_stuff(guitar, 'Am', 2, 160, 20)
    guitar.queue(60./160*22*SAMPLE_RATE, (guitar.strum_up, ('C')))
    guitar.queue(60./160*22.5*SAMPLE_RATE, (guitar.strum_up, ('C')))
    guitar.queue(60./160*23*SAMPLE_RATE, (guitar.strum_down, ('C')))
    buf = guitar[: 60. / 160 * 8].purify()
    buf2 = guitar[60./160*8 : 60./160*16].purify()
    buf3 = guitar[60./160*16 : 60./160*24].purify()

    di = sound.instrument.HardDisk()
    di.tempo = 140
    dn = di.note
    h1 = dn(C3, 2) & dn(D3, 2) & dn(E3) & dn(D3) & dn(C3, 2)
    h2 = h1.src

    si = sound.instrument.Shaker()
    si.tempo = 160
    bi = sound.instrument.BassDrum()
    bi.tempo = 160
    ii = lambda: si.note(0)*10 & si.note(0)*10 & si.note(0)*10 & si.note(0)*10
    pc = lambda: bi.note(40) & si.note(0)*10 & bi.note(40) & si.note(0)*10
    fc = si.note(0)*10 & si.note(0)*10 & si.note(0)*10
    hl = (ii() >> 1)[:60./160*5]

    di.tempo = 160
    h3 = dn(C4) & dn(G3, 0.75) & dn(G3, 0.25) & dn(A3, 0.5) & dn(G3, 0.5) & dn(F3, 0.5) & dn(E3, 0.5) & dn(E3, 0.25) & dn(F3, 0.25) & dn(G3, 1) & dn(C3, 0.5) & dn(D3, 1.5) & dn(C3, 0.5)
    h4 = dn(C4) & dn(G3, 0.75) & dn(G3, 0.25) & dn(A3, 0.5) & dn(G3, 0.5) & dn(F3, 0.5) & dn(E3, 0.5) & dn(C4) & dn(B3) & dn(C4, 2)
    h5 = dn(C4) & dn(E4, 0.75) & dn(C4, 0.25) & dn(F4, 0.5) & dn(E4, 0.5) & dn(D4, 0.5) & dn(C4, 0.5) & dn(D4, 0.25) & dn(E4, 0.25) & dn(D4, 0.25) & dn(E4, 0.25)  & dn(D4, 0.25) & dn(E4, 0.25) & dn(D4, 0.25) & dn(E4, 0.25) & dn(D4, 1.5) & dn(G3, 0.5)
    h6 = dn(C4, 0.5) & dn(B3, 0.5) & dn(A3, 0.5) & dn(G3, 0.5) & dn(F3, 0.5) & dn(E3, 0.5) & dn(D3, 0.5) & dn(C3, 0.5) & dn(C4) & dn(G3) & dn(C4, 2)
    h7 = (buf&buf2&buf2&buf3) / 10 + (h3&h4&h5&h6)*0.8 + (pc()&pc()&pc()&pc()&pc()&pc()&pc()&fc)*0.5
    h8 = h2 & hl & h7

    return p8 & m4 & m6 & n3 & q6 & e5 & f8 & h8

if __name__ == '__main__':
    tune = main_tune()
    if len(sys.argv) > 1:
        tune.write(sys.argv[1])
    else:
        tune.play(progress=True)




# pylint: disable=unused-argument
from .import SAMPLE_RATE
from .filter import HighPassFilter, FakeFMFilter, LowPassFilter, FMFilter, ring_filter, PitchShift
from .envelope import Decay, envelope, Envelope, ADSR
from .tone import SineWave as Sine, harmonics, Noise, Digitar
from .note import Note
from .sample import RawData

__all__ = ('Instrument', 'SineSustain', 'SineHit', 'KickDrum', 'Shaker', 'BassDrum', 'SquareViolin', 'HardDisk', 'ElectricHorn', 'Bell', 'Bell2', 'ElectricBass', 'Guitar')

class Instrument(object):
    """
    A base class for instruments.
    An instrument is an object that can produce notes to a certain specification.
    """
    def __init__(self, tempo=120, volume=1):
        """
        :param tempo:   The tempo to produce notes at, in bpm. Optional, defaults to 120.
        """
        self.tempo = tempo
        self.articulation = 0.8
        self.volume = volume

    @property
    def beat(self):
        # convert from beats/minute to seconds/beat
        # b/m * m/s = b/s
        return 60. / self.tempo

    def rest(self, beats=1):
        """
        Return a rest for the given number of beats.
        """
        return Note(0, beats, self.beat * SAMPLE_RATE)

    def note(self, freq, beats=1, articulation=None):
        """
        Return a note to the given specification.

        :param freq:            The frequency of the note to produce
        :param beats:           The number of beats to produce a note for, default 1
        :param articulation:    An optional number describing how legato or staccato notes should
                                be produced, between 0 and 1. Larger values are more legato, smaller
                                values are more staccato. Doesn't work correctly for precussion or
                                string instruments.
        """
        if articulation is None: articulation = self.articulation
        filling = min(1, articulation / 0.8)
        legato = max(0, (articulation - 0.8) / 0.2)
        n = Note(self._note(freq, beats, filling, legato), beats, self.beat * SAMPLE_RATE) * self.volume
        #n.author = self
        return n

    def _note(self, freq, beats, filling, legato):
        raise NotImplementedError

    def play(self, *args):
        """
        Play a note to the given specification.
        Parameters are the same as those to `note`.
        """
        self.note(*args).play()


class SampledDrum(Instrument):
    def __init__(self, bpm, filenames):
        super(SampledDrum, self).__init__(bpm)
        self.samples = [RawData.from_file(fname) for fname in filenames]

    def _note(self, freq, beats, filling, legato):
        return self.samples[freq]


class SineSustain(Instrument):
    """
    A sustained sine tone.
    """
    def _note(self, freq, beats, filling, legato):
        env = Envelope(beats * self.beat * filling)
        avgenvelope = 0.5*legato*env + (1-legato)*env.apply_adsr(0.01,0.05,0.1, sustain_level=0.5)
        return Sine(freq) * avgenvelope

class SineHit(Instrument):
    """
    A plucked sine tone. Sounds sort of xylophonish.
    """
    def _note(self, freq, beats, filling, legato):
        env = Envelope(beats * self.beat * filling)
        env2 = env.apply_decay(0.095*filling**4 + 0.005)
        assert isinstance(env2, Decay)
        env3 = env2.apply_adsr(0.005, 0, 0.1 - 0.1*legato, sustain_level=1)
        return Sine(freq) * env3

#def xylophone(freq):
#    return Sine(freq/4)    * Decay(0.1, 2, 0.01, 0.3)   * 0.5  + \
#           Sine(freq*8)    * Decay(0.01, 2, 0.01, 0.1)  * 0.15 + \
#           Sine(freq*8*3)  * Decay(0.01, 2, 0.01, 0.2)  * 0.05 + \
#           Sine(50)        * Decay(0.0, 0.2, 0)         * 0.05 + \
#           Sine(60)        * Decay(0.0, 0.2, 0)         * 0.05 + \
#           Sine(70)        * Decay(0.0, 0.2, 0)         * 0.05 + \
#           Sine(80)        * Decay(0.0, 0.2, 0)         * 0.05 + \
#           Sine(90)        * Decay(0.0, 0.2, 0)         * 0.05 + \
#           Sine(100)       * Decay(0.0, 0.1, 0.1)       * 0.05

class KickDrum(Instrument):
    """
    A basic kick drum that's just a low-frequency decaying sine wave.
    Frequency parameters are ignored.
    """
    def _note(self, *args):
        return Sine(110) * Decay(0, 0.4, 0, 0.0000001)

class Shaker(Instrument):
    """
    A basic shaker precussion instrument that's just decaying noise.
    Frequency parameters are ignored.
    """
    def _note(self, *args):
        return (HighPassFilter(Noise(), 0.1) * envelope(decay=0.0001)).purify()

class BassDrum(Instrument):
    """
    A tuned bass drum, works by pitch-shifting a sine wave by a decaying value.
    """
    def _note(self, freq, *args):
        basesound = Sine(freq)
        pitchdecay = envelope(1, decay=0.0001) + 1
        ampdecay = envelope(1, decay=0.0000001)
        return PitchShift(basesound, pitchdecay) * ampdecay

class SquareViolin(Instrument):
    """
    A violin made from a square wave.
    """
    def _note(self, freq, beats, filling, legato):
        #attack = min(0.5, sustain/10.)
        #env = envelope(attack=attack,
        #               decay=0,
        #               sustain=sustain-attack*1.5,
        #               attack_level=attack/2,
        #               sustain_level=1,
        #               decaying_sustain=False)
        env = Envelope(beats * self.beat * filling)
        argenvelope = 0.5*legato*env + (1-legato)*env.apply_adsr(0.01,0.05,0.1, sustain_level=0.5)
        return sum(Sine(freq*i) * 0.5 * 1./i for i in range(1, 5)) * argenvelope

class HardDisk(Instrument):
    """
    An instrument that sounds sort of like when people make music out of hard disks
    """
    def _note(self, freq, beats, filling, legato):
        sample = Digitar(freq, buffersize=2000) << 0.1
        env = Envelope(beats * self.beat * filling)
        return sample * env

class ElectricHorn(Instrument):
    """
    A horn-like sound made with FM synthesis
    """
    def _note(self, freq, beats, filling, legato):
        env = ADSR(0.05, 0.05, beats * self.beat * filling - 0.15, 0.15)
        src = FMFilter(Sine(freq), Sine(freq + 0.5), 30)
        return env * src

class Bell(Instrument):
    """
    A bell sound made from ring modulation. Doesn't sound good at all frequencies.
    """
    def _note(self, freq, *args):
        return envelope(decay=0.1, attack=0.001, attack_level=0.4) * ring_filter(harmonics(freq, (2,3,4)))

class Bell2(Instrument):
    """
    Another bell, sounding more metallic, made from my horrible failed first attempt at FM synthesis.
    Doesn't sound good at all frequencies.
    """
    def _note(self, freq, *args):
        return LowPassFilter(
                envelope(decay=0.3) * FakeFMFilter(Sine, Sine(100. * freq / 220), carrier_freq=freq*2, mod_quantity=50),
                0.1
            )

class ElectricBass(Instrument):
    """
    An electric bass guitar, made with FM synthesis
    I think this one is my favorite.
    """
    def _note(self, freq, beats, filling, legato):
        sineA = Sine(freq)
        sustain = beats * self.beat * filling
        return FMFilter(sineA, Sine(freq/2) * envelope(200)) * envelope(decay=0.1, sustain=sustain)

class Guitar(Instrument):
    """
    A simple guitar made from Karplus-Strong plucked string synthesis
    """
    def _note(self, freq, beats, filling, legato):
        digibase = Digitar(freq)
        return digibase * Envelope(beats * self.beat * filling)



from .signal import ConstantSignal, Signal
from .tone import SineWave

__all__ = ('LowPassFilter', 'BetterLowPassFilter', 'HighPassFilter', 'FakeFMFilter', 'FMFilter', 'ring_filter', 'bessel_wave', 'PitchShift')

numty = (int, float)

class LowPassFilter(Signal):
    """
    Filter the argument signal, cutting out higher-frequency components.

    The beta parameter controls the strength of the filter.
    """
    def __init__(self, src, beta=0.5, pure=False, iir=True):
        self.src = src
        self.beta = beta
        self.pure_param = pure
        self.pure = src.pure and pure
        self.iir = iir
        self.duration = src.duration
        if pure and iir:
            raise Exception("Can't be both pure and infinite")

        self.last_sample = 0
        self.last_frame = -1

    def amplitude(self, frame):
        self.last_frame += 1
        if self.last_frame != frame:
            if self.pure:
                self.last_frame = frame
                self.last_sample = self.src.amplitude(frame-1)
            else:
                raise Exception("Can't access pure samples out of order")

        cur_sample = self.src.amplitude(frame)
        out = cur_sample * self.beta + self.last_sample * (1-self.beta)
        self.last_sample = out if self.iir else cur_sample
        return out

class BetterLowPassFilter(Signal):
    """
    This is a different implementation of the low pass filter.
    Pass it a signal to filter, and then a number of numerical parameters.
    Each frame of the output signal is the linear combination of the given
    parameters and the last n frames from the input signal.
    The last passes parameter corresponds with the most recent input frame.
    """
    def __init__(self, src, *params):
        self.params = list(params)
        eq = float(sum(self.params))
        self.params = [x / eq for x in self.params]
        self.cache = [0.]*len(self.params)
        self.pure = False
        self.duration = src.duration
        self.src = src

    def amplitude(self, frame):
        for i in range(len(self.cache) - 1):
            self.cache[i] = self.cache[i+1]
        self.cache[-1] = self.src.amplitude(frame)

        return sum(x * y for x, y in zip(self.params, self.cache))

class HighPassFilter(Signal):
    """
    Filter the input signal, cutting out low-frequency components.

    The beta parameter controls the strength of the filter.
    """
    def __init__(self, src, beta=0.5):
        self.src = src
        self.beta = beta
        self.pure = False
        self.last_sample = 0
        self.last_out = 0
        self.last_frame = -1
        self.duration = src.duration

    def amplitude(self, frame):
        self.last_frame += 1
        if self.last_frame != frame:
            raise Exception("Out of order access on impure sample: HighPassFilter")

        cur_sample = self.src.amplitude(frame)
        out = self.beta * (self.last_out + cur_sample - self.last_sample)
        self.last_sample = cur_sample
        self.last_out = out
        return out

class FakeFMFilter(Signal):
    """
    Don't use this class please, I am so embarassed
    """
    def __init__(self, carrier_class, modulator, carrier_freq=440, mod_quantity=300):
        self.carrier_class = carrier_class
        self.modulator = modulator
        self.carrier_freq = carrier_freq
        self.mod_quantity = mod_quantity

        sample_carrier = carrier_class(carrier_freq)
        if not sample_carrier.pure:
            raise Exception("Can't use an impure carrier for a FM filter")
        self.duration = min(sample_carrier.duration, modulator.duration)
        self.trigger = 0
        self.trigger_state = True
        self.pure = False # :/

    def _amplitude(self, frame, tframe):
        return self.carrier_class(
                self.carrier_freq + self.mod_quantity * self.modulator.amplitude(frame)
            ).amplitude(tframe)

    def amplitude(self, frame):
        out = self._amplitude(frame, frame - self.trigger)
        if out > 0 and not self.trigger_state:
            if self._amplitude(frame - 1, frame - 1 - self.trigger) <= 0:
                self.trigger = frame
                self.trigger_state = True
        elif out < 0 and self.trigger_state:
            if self._amplitude(frame - 1, frame - 1 - self.trigger) >= 0:
                self.trigger_state = False

        return out

class FMFilter(Signal):
    """
    Modulate the frequency of the carrier signal with the modulator signal.
    The strength of the modulator may be adjusted by the mod_quantity parameter.
    """
    def __init__(self, carrier, modulator, mod_quantity=300):
        if not carrier.pure:
            raise Exception("Can't use an impure carrier for an FM filter")
        self.carrier = carrier
        self.modulator = modulator
        self.mod_quantity = mod_quantity
        self.duration = carrier.duration
        self.pure = modulator.pure

    def amplitude(self, frame):
        return self.carrier.amplitude(frame + self.mod_quantity * self.modulator.amplitude(frame))

def ring_filter(data):
    """
    Perform ring modulation on a given number of signals.

    :param data:    A list of signals to modulate together
    """
    out = 1
    for src in data:
        out *= src
    return out

def bessel_wave(freq, alpha, beta):
    """
    I literally don't care if this isn't technically a thing

    Outputs an FM-synthesized sample with carrier frequency freq, modulator frequency
    freq * alpha, and modulator intensity beta
    """
    return FMFilter(SineWave(freq), SineWave(alpha), beta)


class PitchShift(Signal):
    """
    Perform a pitch shift on the source signal by the shift signal.

    The value of the shift signal serves as a multiplier for the frequency of the source.
    Note that this works just by playing the source faster, so putting this on top of an
    enveloped sound is a bad idea, you need to put this under the envelope.
    """
    def __init__(self, src, shift):
        if isinstance(shift, numty):
            shift = ConstantSignal(shift)
        self.src = src
        self.shift = shift - 1
        self.phase = 0

        self.pure = False
        self.duration = src.duration

        self.done = False

    def amplitude(self, frame):
        self.phase += self.shift.amplitude(frame)
        intpart = int(self.phase)
        fracpart = self.phase - intpart
        f1 = frame + intpart
        f2 = f1 + 1
        if f2 >= self.duration:
            self.done = True

        s1 = self.src.amplitude(f1)
        s2 = self.src.amplitude(f2)
        return s1 * (1-fracpart) + s2 * fracpart


class Convolution(Signal):
    def __init__(self, a, b):
        self.a = a
        self.b = b
        self.duration = a.duration + b.duration

    def amplitude(self, frame):
        out = 0.
        f = frame
        g = 0
        while f > 0 and g < self.b.duration:
            out += self.a.amplitude(f) * self.b.amplitude(g)
            f -= 1
            g += 1

        return g



import wave
import struct

from . import SAMPLE_RATE, sd
from .signal import Signal

class RawData(Signal):
    """
    A signal which is specified by an array of sample values.
    """
    def __init__(self, data):
        self.data = data
        self.duration = len(data)

    def amplitude(self, frame):
        if frame < 0: return 0
        try:
            return self.data[frame]
        except IndexError:
            return 0

    @staticmethod
    def from_file(filename):
        """
        Make a new RawData by loading from a .wav file.
        """
        fp = wave.open(filename)
        n = fp.getnframes()
        bdata = fp.readframes(n)
        idata = struct.unpack('h'*n, bdata)
        fdata = [float(x)/(2**15 - 1) for x in idata]
        return RawData(fdata)

    @staticmethod
    def record(seconds):
        """
        Make a new RawData by recording from the microphone.
        """
        data = sd.rec(seconds * SAMPLE_RATE, blocking=True)
        return RawData(data)



import re

__all__ = ['note', 'notename', 'scales', 'A0', 'Ab0', 'As0', 'B0', 'Bb0', 'Bs0', 'C0', 'Cb0', 'Cs0', 'D0', 'Db0', 'Ds0', 'E0', 'Eb0', 'Es0', 'F0', 'Fb0', 'Fs0', 'G0', 'Gb0', 'Gs0', 'A1', 'Ab1', 'As1', 'B1', 'Bb1', 'Bs1', 'C1', 'Cb1', 'Cs1', 'D1', 'Db1', 'Ds1', 'E1', 'Eb1', 'Es1', 'F1', 'Fb1', 'Fs1', 'G1', 'Gb1', 'Gs1', 'A2', 'Ab2', 'As2', 'B2', 'Bb2', 'Bs2', 'C2', 'Cb2', 'Cs2', 'D2', 'Db2', 'Ds2', 'E2', 'Eb2', 'Es2', 'F2', 'Fb2', 'Fs2', 'G2', 'Gb2', 'Gs2', 'A3', 'Ab3', 'As3', 'B3', 'Bb3', 'Bs3', 'C3', 'Cb3', 'Cs3', 'D3', 'Db3', 'Ds3', 'E3', 'Eb3', 'Es3', 'F3', 'Fb3', 'Fs3', 'G3', 'Gb3', 'Gs3', 'A4', 'Ab4', 'As4', 'B4', 'Bb4', 'Bs4', 'C4', 'Cb4', 'Cs4', 'D4', 'Db4', 'Ds4', 'E4', 'Eb4', 'Es4', 'F4', 'Fb4', 'Fs4', 'G4', 'Gb4', 'Gs4', 'A5', 'Ab5', 'As5', 'B5', 'Bb5', 'Bs5', 'C5', 'Cb5', 'Cs5', 'D5', 'Db5', 'Ds5', 'E5', 'Eb5', 'Es5', 'F5', 'Fb5', 'Fs5', 'G5', 'Gb5', 'Gs5', 'A6', 'Ab6', 'As6', 'B6', 'Bb6', 'Bs6', 'C6', 'Cb6', 'Cs6', 'D6', 'Db6', 'Ds6', 'E6', 'Eb6', 'Es6', 'F6', 'Fb6', 'Fs6', 'G6', 'Gb6', 'Gs6', 'A7', 'Ab7', 'As7', 'B7', 'Bb7', 'Bs7', 'C7', 'Cb7', 'Cs7', 'D7', 'Db7', 'Ds7', 'E7', 'Eb7', 'Es7', 'F7', 'Fb7', 'Fs7', 'G7', 'Gb7', 'Gs7', 'A8', 'Ab8', 'As8', 'B8', 'Bb8', 'Bs8', 'C8', 'Cb8', 'Cs8', 'D8', 'Db8', 'Ds8', 'E8', 'Eb8', 'Es8', 'F8', 'Fb8', 'Fs8', 'G8', 'Gb8', 'Gs8', 'A9', 'Ab9', 'As9', 'B9', 'Bb9', 'Bs9', 'C9', 'Cb9', 'Cs9', 'D9', 'Db9', 'Ds9', 'E9', 'Eb9', 'Es9', 'F9', 'Fb9', 'Fs9', 'G9', 'Gb9', 'Gs9'] # type: ignore[reportUnsupportedDunderAll]

scales = {
    'major': [0, 2, 4, 5, 7, 9, 11],
    'minor': [0, 2, 3, 5, 7, 9, 10],
    'chromatic': [0,1,2,3,4,5,6,7,8,9,10,11]
}

def note(degree, octave=3, key='C', scale='major', accidental=''):
    """
    Return the frequency of the note with the given characteristics

    :param degree:      The scale degree of the note, zero-based
    :param octave:      The octave of the note
    :param key:         The key of the scale
    :param scale:       The name of the scale as a string.
                        Options are "major", "minor", and "chromatic".
    :param accidental:  The accidental for this note as a string.
                        Options are "sharp", "#", or "s" for sharp,
                        "flat", "b", or "f" for flat.

    Note that accidentals do not respect the key - if you are in G major
    and you ask for the sixth degree (ti) sharpened, you will get back G,
    not F# like you would play if you were reading sheet music.
    Consequentially, there is no natural accidental.
    """
    pianokey = (octave*12 + scales[scale][degree % len(scales[scale])])

    if accidental.lower() in ('sharp', '#', 's'):
        pianokey += 1
    elif accidental.lower() in ('flat', 'b', 'f'):
        pianokey -= 1
    elif accidental.lower() in ('doublesharp', '##', 'ss'):
        pianokey += 2
    elif accidental.lower() in ('doubleflat', 'bb', 'ff'):
        pianokey -= 2

    return 2**(pianokey/12.) * notename(key + '0')

def notename(name):
    """
    Translate the name of a note, like "A4", into its frequency in hertz.
    """
    if name == 'C0':
        return 27.5 * 2**(-9./12)
    m = re.search('([A-G])(|b|#|s|f)([0-9])', name)
    if m is None:
        raise ValueError(f"Invalid note {name}")
    notestr, accidental, octave = m.group(1), m.group(2), m.group(3)
    degree = {'C': 0, 'D': 1, 'E': 2, 'F': 3, 'G': 4, 'A': 5, 'B': 6}[notestr]
    octavenum = int(octave)
    return note(degree, octavenum, accidental=accidental, key='C', scale='major')

for _octave in range(10):
    for _letter in 'ABCDEFG':
        for _accidental in ['', 'b', 's']:
            _name = _letter + _accidental + str(_octave)
            vars()[_name] = notename(_name)



from . import SAMPLE_RATE
from .signal import ConstantSignal, Signal

__all__ = ('Note',)

numty = (int, float)

class Note(Signal):
    """
    An abstraction over the signal base class to add timing information.
    The idea is that a note has a signal and then the value that that signal should be
    used as, in beats, in addition to the length of a beat.

    All the binary operators on Sample objects now work with beats instead of seconds.
    """
    def __init__(self, src, value, beat):
        """
        :param src:     The signal to use as a source for the note
        :param value:   The value of the note in beats
        :param beat:    The length of a beat in samples
        """
        self.src = ConstantSignal.wrap(src)
        self.value = value
        self.beat = beat
        self.duration = self.src.duration
        self.pure = self.src.pure

    def amplitude(self, frame):
        return self.src.amplitude(frame)

    def __add__(self, other):
        if other == 0:
            return self
        return Note(self.src + other.src, max(self.value, other.value), other.beat)

    def __mul__(self, other):
        if type(other) in numty:
            return Note(self.src * other, self.value, self.beat)
        if type(other) is not Note:
            return Note(self.src * other, self.value, self.beat)
        return Note(self.src * other.src, max(self.value, other.value), other.beat)

    def __neg__(self):
        return Note(-self.src, self.value, self.beat)

    def __rshift__(self, other):
        if type(other) not in numty:
            raise TypeError("Can't shift by %s" % repr(other))
        realshift = other * self.beat / SAMPLE_RATE
        return Note(self.src >> realshift, self.value + other, self.beat)

    def __and__(self, other):
        return self + (other >> self.value)

    def __mod__(self, other):
        if type(other) not in numty:
            raise TypeError("Can't loop by %s" % repr(other))
        if other == 0:
            other = self.value
        return Note(self.src % (other * self.beat / SAMPLE_RATE), float('inf'), self.beat)



# pylint: disable=superfluous-parens
from typing import Optional
import numpy
import struct
import wave

import progressbar

from . import SAMPLE_RATE, sd

numty = (int, float)

__all__ = ('Signal', 'LoopSignal', 'DelaySignal', 'SequenceSignal', 'InvertSignal', 'ConstantSignal', 'MixSignal', 'EnvelopeSignal', 'Purifier', 'SliceSignal', 'ReverseSignal')

class Signal(object):
    """
    The base class for all signal objects. Represents the abstract concept of a signal over time,
    sampled at the given sample rate.

    :ivar pure:     A bool describing whether this signal is pure or not - a pure signal carries
                    no internal state and any frame of it may be accessed in O(1) time.
    :ivar duration: The length of this signal, in frames.

    Several binary operators are overloaded so you may use them on Signal objects.
    A listing that ends with "Constants ok" means that you can use a contant integer or float
    where a signal is expected and it will be converted automatically to an infinitely long
    constant signal of that value.

    - ``+``: You may add signals together to mix them together. Constants ok.
    - ``-``: You may subtract signals and it's like adding one to the inverse of the other.
      You may also negate a signal. Constants ok.
    - ``*``: You may multiply two signals together to perform enveloping or amplitude modulation.
      Constants ok.
    - ``/``: You may divide a signal by a number to reduce its amplitude by that factor.
    - ``>>``: You may right-shift a signal by a number to delay it by that number of seconds.
    - ``<<``: You may left-shift a signal by a number to move it back in time by that number of seconds.
    - ``&``: You may and two signals together to concatenate them.
    - ``%``: You may modulate a signal by a number to loop the first n seconds of it.

    Additionally, you may use array slice notation to extract slices of sample data.
    The slice bounds are in seconds. Normal array indexing does not do anything.
    """
    # pylint: disable=unused-argument,no-self-use

    def play(self, length: Optional[float]=None, progress=False):
        """
        Play this signal. Block until playback is complete.
        If the given signal is infinitely long, default to three seconds of playback.

        :param length:      The length to play, in seconds. Optional.
        :param progress:    Whether to show a progress bar for rendering
        """
        data = self.render(length, progress)
        sd.play(data, blocking=True)

    def play_async(self):
        """
        Play this signal asynchronously. Return the `sounddevice` stream object for this playback.
        The only way you should ever really have to interact with the return value of this function is
        to call `.stop()` on it.

        :param thing:       The signal to play
        """
        def cb(outdata, frames, time, status):  # pylint: disable=unused-argument
            nonlocal timer
            startframe = timer
            for i in range(frames):
                outdata[i] = self.amplitude(i+startframe)
            timer += frames
            if timer >= self.duration:
                raise sd.CallbackStop

        stream = sd.OutputStream(callback=cb)
        timer = 0
        stream.start()
        return stream

    def write(self, filename, length=None, progress=True):
        """
        Write this signal to a .wav file.

        :param filename:    The filename to write to. Regardless of its extension, the output filetype
                            will be uncompressed .wav
        :param thing:       The signal to write
        """
        data = self.render(length, progress)
        mm = 2**15 - 1
        fp = wave.open(filename, 'w')
        fp.setparams((1, 2, 44100, 63822, 'NONE', 'not compressed'))
        fp.writeframes(b''.join(struct.pack('h', int(max(min(dat, 1), -1)*mm)) for dat in data))
        fp.close()

    def render(self, length=None, progress=False, clip_warn=True):
        """
        Render this signal into an numpy array of floats. Return the array.

        :param length:      The length to render, in seconds. Optional.
        :param progress:    Whether to show a progress bar for rendering
        """
        if progress and not progressbar:
            print('Install the progressbar module to see a progress bar for rendering')
            progress = False

        duration = self.duration if length is None else length * SAMPLE_RATE
        if duration == float('inf'):
            duration = 3*SAMPLE_RATE
        else:
            duration = int(duration)
        out = numpy.empty((duration,))

        pbar = progressbar.ProgressBar(widgets=['Rendering: ', progressbar.Percentage(), ' ', progressbar.Bar(), ' ', progressbar.ETA()], maxval=duration-1).start() if progress else None

        clipped = 0
        for i in range(duration):
            out[i] = self.amplitude(i)
            if pbar: pbar.update(i)
            if clip_warn and abs(out[i]) > 1:
                clipped = max(abs(out[i]), clipped)
                out[i] = 1 if out[i] > 0 else -1
        if pbar: pbar.finish()
        if clip_warn and clipped != 0:
            print('Warning: clipping! max val %s' % clipped)
        return out

    def amplitude(self, frame: int) -> float:
        """
        The main interface for accessing sample data. This is the primary method that should
        be overridden by subclasses.

        :param frame:       The frame whose amplitude should be returned.
        """
        return 0.

    duration = 0
    pure = True

    def __add__(self, other) -> "Signal":
        if other == 0:
            return self
        if type(other) is DelaySignal:
            return SequenceSignal((self, 0), (other.src, other.delay))
        return MixSignal(self, ConstantSignal.wrap(other))

    def __radd__(self, other) -> "Signal":
        return self + other

    def __sub__(self, other) -> "Signal":
        if other == 0:
            return self
        return self + -ConstantSignal.wrap(other)

    def __rsub__(self, other) -> "Signal":
        return -self + ConstantSignal.wrap(other)

    def __mul__(self, other) -> "Signal":
        if other == 1:
            return self
        if other == 0 or (type(other) == ConstantSignal and other._amplitude == 0):
            return ConstantSignal(0)
        return EnvelopeSignal(self, ConstantSignal.wrap(other))

    def __rmul__(self, other) -> "Signal":
        return self * other

    def __div__(self, other) -> "Signal":
        if type(other) not in numty:
            raise TypeError("Can't divide by %s" % repr(other))
        return self * (1./other)

    __floordiv__ = __div__
    __truediv__ = __div__

    def __rdiv__(self, other):
        raise TypeError("Can't divide by %s" % repr(self))

    def __neg__(self) -> "Signal":
        return InvertSignal(self)

    def __rshift__(self, other) -> "Signal":
        if type(other) not in numty:
            raise TypeError("Can't shift by %s" % repr(other))
        return DelaySignal(self, int(other*SAMPLE_RATE))

    def __lshift__(self, other) -> "Signal":
        return self >> -other

    def __and__(self, other) -> "Signal":
        return self + (other >> (float(self.duration) / SAMPLE_RATE))

    def __mod__(self, other) -> "Signal":
        if type(other) not in numty:
            raise TypeError("Can't loop by %s" % repr(other))
        return LoopSignal(self, other)

    def __getitem__(self, key):
        if isinstance(key, slice):
            if key.step is not None:
                raise KeyError(key)
            start = 0 if key.start is None else key.start
            stop = float(self.duration)/SAMPLE_RATE if key.stop is None else key.stop
            return SliceSignal(self, start, stop)
        else:
            raise KeyError(key)

    def purify(self, preprocess=False) -> "Signal":
        """
        Return a pure version of this signal. This is a no-op for pure signals, but for
        impure signals it installs a caching layer on top of the signal.

        :param preprocess:      Whether the cache should preload the sample data at initialize-time.
                                Optional.
        """
        if not preprocess and self.pure:
            return self
        return Purifier(self, preprocess=preprocess)

    def reverse(self) -> "Signal":
        """
        Return a reversed version of this signal.
        """
        return ReverseSignal(self)

class LoopSignal(Signal):
    """
    A signal that loops the first n seconds of its child
    """
    def __init__(self, src, length):
        self.src = src.purify()     # lmao
        self.length = int(length * SAMPLE_RATE) if length != 0 else src.duration
        self.duration = float('inf')
        self.pure = True

    def amplitude(self, frame):
        cur_frame = frame % self.length
        out = 0.

        while frame >= 0 and cur_frame < self.src.duration:
            out += self.src.amplitude(cur_frame)
            frame -= self.length
            cur_frame += self.length

        return out

class DelaySignal(Signal):
    """
    A signal that delays its child by n seconds
    """
    def __init__(self, src, delay):
        """
        delay is in samples
        """
        self.src = src
        self.delay = delay
        self.duration = src.duration + int(self.delay)
        self.pure = src.pure

    def amplitude(self, frame):
        return self.src.amplitude(frame - self.delay)

    def __rshift__(self, other):
        if type(other) not in numty:
            raise TypeError("Can't shift by %s" % repr(other))
        return DelaySignal(self.src, self.delay + int(other*SAMPLE_RATE))

    def __add__(self, other):
        if type(other) is DelaySignal:
            return SequenceSignal((self.src, self.delay), (other.src, other.delay))
        else:
            return super(DelaySignal, self).__add__(other)

class SequenceSignal(Signal):
    """
    A sequence of signals starting at specific points in time.
    Ultimately used as an optimization for combinations of the `>>`, `&`, and `+` operators.
    """
    def __init__(self, *data):
        """
        data is a sequence of tuples of (Signal, starttime)
        starttime is in samples
        """
        if len(data) == 1 and hasattr(data[0], '__iter__'):
            data = data[0]
        data = list(data)
        assert all(type(x[1]) in numty for x in data)
        self.srcs = sorted(((src, start, start + src.duration) for src, start in data), key=lambda x: x[1])
        self.duration = max(src[2] for src in self.srcs)
        self.pure = all(src[0].pure for src in self.srcs)

    def amplitude(self, frame):
        # TODO: optimize this maybe?
        out = 0.
        for src, start, end in self.srcs:
            if frame >= start and frame < end:
                out += src.amplitude(frame - start)
        return out

    def __rshift__(self, other):
        if type(other) not in numty:
            raise TypeError("Can't shift by %s" % repr(other))
        return SequenceSignal((src, start + int(other*SAMPLE_RATE)) for (src, start, _) in self.srcs)

    def __add__(self, other):
        if type(other) is DelaySignal:
            return SequenceSignal((other.src, other.delay), *((src, start) for (src, start, _) in self.srcs))
        elif type(other) is SequenceSignal:
            return SequenceSignal((src, start) for src, start, _ in self.srcs + other.srcs)
        else:
            return MixSignal(self, ConstantSignal.wrap(other))

class InvertSignal(Signal):
    """
    A signal that inverts its child
    """
    def __init__(self, src):
        self.src = src
        self.duration = src.duration
        self.pure = src.pure

    def amplitude(self, frame):
        return -self.src.amplitude(frame)

    def __neg__(self):
        return self.src

class ConstantSignal(Signal):
    """
    A signal that is a constant value
    """
    def __init__(self, amplitude):
        self._amplitude = amplitude
        self.duration = 0 if amplitude == 0 else float('inf')
        self.pure = True

    def amplitude(self, frame):
        return self._amplitude

    @staticmethod
    def wrap(val):
        if type(val) in numty:
            return ConstantSignal(val)
        return val

class MixSignal(Signal):
    """
    A signal that mixes all its children together
    """
    def __init__(self, *signals):
        if len(signals) == 1 and hasattr(signals[0], '__iter__'):
            signals = signals[0]
        self.signals = list(signals)
        self.pure = all(s.pure for s in self.signals)
        try:
            self.duration = max(s.duration for s in self.signals if s.duration != float('inf'))
        except ValueError:
            self.duration = float('inf')

    def amplitude(self, frame):
        return sum(s.amplitude(frame) for s in self.signals)

    def __add__(self, other):
        if type(other) is MixSignal:
            return MixSignal(self.signals + other.signals)
        return MixSignal(ConstantSignal.wrap(other), *self.signals)

class EnvelopeSignal(Signal):
    """
    A signal that implements enveloping and amplitude modulation
    """
    def __init__(self, src, envelope):
        self.src = src
        self.env = envelope
        self.duration = min(self.env.duration, self.src.duration)
        self.pure = self.src.pure and self.env.pure

    def amplitude(self, frame):
        return self.src.amplitude(frame)*self.env.amplitude(frame)

class Purifier(Signal):
    """
    A signal that caches its child's amplitude data
    """
    def __init__(self, src, length=None, preprocess=False):
        if length is None:
            if src.duration == float('inf') and preprocess:
                raise ValueError("Cannot purify an infinite number of samples")
            length = src.duration
        else:
            length = int(length * SAMPLE_RATE)
        self.nextf = 0
        self.duration = length
        self.storage = [0.]*(100000 if self.duration == float('inf') else int(self.duration))
        self.pure = True
        self.src = src

        if preprocess:
            self.amplitude(self.duration - 1)

    def amplitude(self, frame):
        if frame < 0: return 0.
        if frame >= self.duration: return 0.
        while frame >= self.nextf:
            if self.nextf >= len(self.storage):
                self.storage += [0.]*100000
            self.storage[self.nextf] = self.src.amplitude(self.nextf)
            self.nextf += 1
        return self.storage[frame]

class SliceSignal(Signal):
    """
    A signal that extracts a slice of its child
    """
    def __init__(self, src, from_time, to_time, relative=False):
        self.from_frame = int(from_time * SAMPLE_RATE)
        try:
            self.to_frame = int(to_time * SAMPLE_RATE)
        except OverflowError:
            self.to_frame = float('inf')

        if relative:
            self.to_frame += self.from_frame
        self.duration = self.to_frame - self.from_frame
        self.pure = src.pure
        self.src = src

    def amplitude(self, frame):
        if frame < 0: return 0
        if frame >= self.duration: return 0
        return self.src.amplitude(frame + self.from_frame)

class ReverseSignal(Signal):
    """
    A signal that reverses its child
    """
    def __init__(self, src):
        self.src = src
        self.duration = src.duration

    def amplitude(self, frame):
        return self.src.amplitude(self.duration - frame - 1)

    def reverse(self):
        return self.src



import math
import random

from . import SAMPLE_RATE
from .signal import Signal

__all__ = ('Sample', 'SineWave', 'SquareWave', 'SawtoothWave', 'TriangleWave', 'Noise', 'BrownNoise', 'Digitar', 'harmonics', 'wavetable')

class Sample(Signal):
    """
    An infinitely long sound generated by a simple algorithm
    """
    def __init__(self, frequency):
        self.frequency = float(frequency)
        self.duration = float('inf')

    @property
    def period(self):
        return 1/self.frequency * SAMPLE_RATE

class SineWave(Sample):
    """
    A sample that outputs a sine wave at the given frequency
    """
    def amplitude(self, frame):
        return math.sin(frame * 2 * math.pi / self.period)

class SquareWave(Sample):
    """
    A sample that outputs a square wave at the given frequency

    Due to aliasing, you may want to use a series of sine wave harmonics if you want to
    obtain a more pleasant sound.
    """
    def __init__(self, frequency, split=0.5):
        super(SquareWave, self).__init__(frequency)
        self.split = split

    def amplitude(self, frame):
        return 1 if frame % self.period < self.period*self.split else -1

class SawtoothWave(Sample):
    """
    A sample that outputs a sawtooth wave at the given frequency

    Due to aliasing, you may want to use a series of sine wave harmonics if you want to
    obtain a more pleasant sound.
    """
    def amplitude(self, frame):
        return frame % self.period / self.period * 2 - 1

class TriangleWave(Sample):
    """
    A sample that outputs a triangle wave at the given frequency

    Due to aliasing, you may want to use a series of sine wave harmonics if you want to
    obtain a more pleasant sound.
    """
    def amplitude(self, frame):
        pframe = frame % self.period
        hperiod = self.period/2
        qperiod = hperiod/2
        if pframe < qperiod:
            return pframe / qperiod
        pframe -= qperiod
        if pframe < hperiod:
            return pframe / -hperiod*2 + 1
        pframe -= hperiod
        return pframe / qperiod - 1

class Noise(Sample):
    """
    A sample that outputs white noise, random data uniformly distributed over [0,1].
    """
    # I... guess this is technically pure?
    def __init__(self):
        super(Noise, self).__init__(0)

    def amplitude(self, frame):
        return random.random() * 2 - 1

class BrownNoise(Sample):
    """
    A sample that outputs brown noise, the integration of white noise.

    It is technically pure, but output may more closely resemble white noise if sample impurely.
    """
    def __init__(self, fac=0.5):
        super(BrownNoise, self).__init__(0)
        self.prev = 0.
        self.fac = float(fac)

    def amplitude(self, frame):
        self.prev += (random.random() * 2 - 1) * self.fac
        if self.prev > 1: self.prev = 1.
        elif self.prev < -1: self.prev = -1.
        return self.prev


def wavetable(table):
    class WaveSample(Sample):
        def __init__(self, freq):
            Sample.__init__(self, freq)
            basefreq = SAMPLE_RATE * 1./len(table)
            self.phaseinc = freq / basefreq

        def amplitude(self, frame):
            realframe = frame * self.phaseinc
            s1 = table[int(realframe) % len(table)]
            s2 = table[int(realframe + 1) % len(table)]
            interp = realframe - int(realframe)
            return s2 * interp + s1 * (1 - interp)

    return WaveSample


class Digitar(Sample):
    """
    A sample that implements the Karplus-Strong plucked string synthesis algorithm.
    The basic idea is that an wavetable initially populated with random noise run though
    a low-pass filter cyclically, gradually removing inharmonic components and smoothing
    the waveform. The sound is tuned by keeping a separate "phase" counter which causes the
    output to cycle through the wavetable at a different speed than the filter, effectively
    adjusting the period of the signal. The decay rate is adjusted by changing the size of
    the wavetable - the smaller the wavetable, the faster the filter adjusts the signal and
    the faster that the sound decays. Keep in mind that if it decays very quickly, then
    the noise doesn't have time to resolve into a tone, and the output will sound more
    drum-like.

    :param frequency:       The desired output frequency
    :param buffersize:      The size of the wavetable.
                            Optional, defaults to a good plucked string sound.
    :param wavesrc:         A signal to sample to produce the initial wavetable.
                            Optional, defaults to white noise.
    """
    def __init__(self, frequency, buffersize=256, wavesrc=None):
        self.wavesrc = wavesrc if wavesrc is not None else Noise()
        super(Digitar, self).__init__(frequency)
        self.buffersize = buffersize
        self.sample_window = []
        self.cur_frame = 0
        basefreq = SAMPLE_RATE * 1./self.buffersize
        self.phaseinc = frequency / basefreq
        self.phase = 0
        self.new_buffer()
        self.pure = False

    def new_buffer(self):
        self.sample_window = [self.wavesrc.amplitude(i) for i in range(self.buffersize)]
        self.cur_frame = 0
        self.phase = 0

    def get_buffer(self, frame):
        return self.sample_window[frame % self.buffersize]

    def set_buffer(self, frame, value):
        self.sample_window[frame % self.buffersize] = value

    def tick(self):
        self.set_buffer(self.cur_frame + 1, self.get_buffer(self.cur_frame) * 0.3 + self.get_buffer(self.cur_frame + 1) * 0.7)
        self.cur_frame += 1
        self.phase = (self.phase + self.phaseinc) % self.buffersize

    def seek(self, frame):
        if frame < self.cur_frame:
            self.cur_frame = 0
        while self.cur_frame < frame:
            self.tick()

    def amplitude(self, frame):
        self.seek(frame)
        s1 = self.get_buffer(int(self.phase))
        s2 = self.get_buffer(int(self.phase) + 1)
        interp = self.phase - int(self.phase)
        return interp * s2 + (1-interp) * s1

def harmonics(freq, ns=(1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16), subsample=SineWave):
    """
    Generate a number of harmonics of a given sample. The base tone is treated as the first harmonic.

    :param freq:        The base frequency to use
    :param ns:          A list of the harmonics to produce.
                        Optional, defauts to the first 16 harmonics.
    :param subsample:   The class of the sample to use.
                        Optional, defaults to a sine wave.
    """
    return [subsample(freq*n) for n in ns]



# This is a parser for json files that describe melodies
# the provided format.json provides two sample melodies
# the format can define its own instruments

# pylint: disable=unnecessary-lambda,abstract-method

import sound
import sys
import json
import copy
from functools import reduce

try:
    num = (int, long, float)
except NameError:
    num = (int, float)
try:
    anystr = (str, unicode)
except NameError:
    anystr = (str, bytes)

Sig = sound.tone.Signal
available_signals = [
        (lambda sine: sound.tone.SineWave(sine), (num,)),
        (lambda square: sound.tone.SquareWave(square), (num,)),
        (lambda triangle: sound.tone.TriangleWave(triangle), (num,)),
        (lambda sawtooth: sound.tone.SawtoothWave(sawtooth), (num,)),
        (lambda add_1, add_2: add_1 + add_2, (num, num)),
        (lambda sub_1, sub_2: sub_1 - sub_2, (num, num)),
        (lambda mul_1, mul_2: mul_1 * mul_2, (num, num)),
        (lambda div_1, div_2: div_1 / div_2, (num, num)),
        (lambda volume, source: volume * source, (num, Sig)),
        (lambda envelope, envelope_source: envelope * envelope_source, (Sig, Sig)),
        (lambda adsr_attack, adsr_decay, adsr_sustain, adsr_release, adsr_level: sound.envelope.ADSR(adsr_attack, adsr_decay, adsr_sustain, adsr_release, sustain_level=adsr_level), (num,)*5),
        (lambda decay_speed, decay_length: sound.envelope.Decay(0, decay_length, 0.01, 1./decay_speed), (num, num)),
        (lambda fm_carrier, fm_modulator, fm_quantity: sound.filter.FMFilter(fm_carrier, fm_modulator, fm_quantity), (Sig, Sig, num))
]

class UserInstrument(sound.instrument.Instrument):
    def __init__(self, name, ast, tempo=120):
        super(UserInstrument, self).__init__(tempo)
        self.name = name
        self.ast = ast

    def note(self, freq, beats=1, articulation=None):
        if articulation is None: articulation = self.articulation
        sig = self.parse_signal(self.name, self.ast, {'$FREQUENCY': freq, '$DURATION': beats * self.beat, '$ARTICULATION': articulation})
        if sig.duration == float('inf'):
            sig.duration = beats * self.beat * sound.SAMPLE_RATE
        return sound.note.Note(sig, beats, self.beat * sound.SAMPLE_RATE)

    @staticmethod
    def typecheck(name, signature, args):
        if len(signature) != len(args):
            raise TypeError("%s: need %d args, got %d" % (name, len(signature), len(args)))

        for i, (ty, arg) in enumerate(zip(signature, args)):
            if not isinstance(arg, ty):
                raise TypeError("%s, arg %d: expected %s, got %s" % (name, i, ty, type(arg)))

    @staticmethod
    def parse_helper(name, func, types, item, env):
        keys = func.__code__.co_varnames
        assert len(keys) == len(types)
        for key in keys:
            if key not in item:
                return None
        args = [UserInstrument.parse_signal(name + ' > ' + key, item[key], env) for key in keys]
        return func(*args)

    @staticmethod
    def parse_signal(name, signal, env):
        if type(signal) in num:
            return signal
        elif type(signal) in anystr:
            if signal in env:
                return env[signal]
            else:
                raise ValueError("%s: Bad identifier: %s" % (name, signal))
        elif type(signal) is dict:
            for func, types in available_signals:
                val = UserInstrument.parse_helper(name, func, types, signal, env)
                if val is not None:
                    return val
            raise ValueError("%s: Unknown function!" % name)
        else:
            raise TypeError("%s: VERY bad type" % name)


class MusicData(object):
    def __init__(self, filename):
        try:
            self.json_info = json.load(open(filename))
        except:
            raise ValueError("file %s either does not exist or is not valid JSON" % filename)

        if not 'melodies' in self.json_info:
            raise ValueError("file must contain 'melodies' key at top level")

        self.default_instruments = self.json_info.get('default_instruments', {})
        self.default_instrument = 'sine'
        self.instruments = {key: UserInstrument(key, val) for key, val in self.json_info.get('instruments', {}).items()}
        self.instruments['sine'] = UserInstrument("sine", { "sine": "$FREQUENCY" })

        self.tempo = self.json_info.get('tempo', 120)
        if type(self.tempo) not in (int, float):
            raise TypeError('%s is not a valid tempo' % repr(self.tempo))

    @property
    def melodies(self):
        return [self.melody(name) for name in self.json_info['melodies']]

    def melody(self, name, instrument=None):
        if not name in self.json_info['melodies']:
            raise KeyError(name)

        if instrument is None:
            instrument = self.default_instruments.get(name, self.default_instrument)
        if instrument in self.instruments:
            instrument = self.instruments[instrument]
        else:
            raise ValueError("No such instrument: %s" % instrument)
        instrument.tempo = self.tempo
        instrument.volume = 0.1

        melody = self.json_info['melodies'][name]
        if type(melody) is not list:
            raise TypeError('melody %s is not a list' % name)

        return self._parse_melody(melody, instrument)

    def _parse_melody(self, sequence, instrument):
        out = None

        for note in sequence:
            inst = copy.copy(instrument)
            p = None

            if "articulation" in note:
                inst.articulation = note['articulation']
            if "volume" in note:
                inst.volume = note['volume'] * 0.01
            if "tempo" in note:
                inst.tempo = note['tempo']

            if "name" in note and 'duration' in note:
                p = inst.note(sound.notes.notename(note['name']), note['duration'])
            elif "rest" in note:
                p = inst.rest(note['rest'])
            elif "chord" in note and "duration" in note:
                p = sum(inst.note(sound.notes.notename(x), note['duration']) for x in note['chord']) / len(note['chord'])
            elif "scope" in note:
                if note['scope'] == 'persist':
                    instrument = inst
                elif type(note['scope'] is list):
                    p = self._parse_melody(note['scope'], inst)
            else:
                raise TypeError('note %s is of no known form!' % note)

            if "loop" in note:
                if p.pure:
                    p = reduce(lambda x, y: x & y, [p]*note['loop'])
                else:
                    ct = note['loop'] - 1
                    cpy = copy.copy(note)
                    del cpy['loop']
                    p &= self._parse_melody([cpy]*ct, inst)

            if out is None:
                out = p
            else:
                out &= p

        return out


def main(filename, melodyname):
    music = MusicData(filename)
    melody = music.melody(melodyname)
    melody.play()
    #with sound.play_async(melody):
    #    time.sleep(melody.duration / sound.SAMPLE_RATE)

def usage(argv0):
    print('Usage: %s filename melodyname' % argv0)

if __name__ == '__main__':
    if len(sys.argv) > 2:
        main(sys.argv[1], sys.argv[2])
    else:
        usage(sys.argv[0])



# pylint: disable=redefined-builtin
import sounddevice as sd
SAMPLE_RATE = 44100
sd.default.samplerate = SAMPLE_RATE  # type: ignore
sd.default.channels = 1  # type: ignore

__all__ = ('tone', 'sample', 'envelope', 'filter', 'instrument', 'notes', 'note', 'asyncplayer')
__version__ = "1.1.0"
released_version = "1.1.0"

from . import tone, sample, envelope, filter, instrument, notes, note, asyncplayer




from typing import Union
import math

from . import SAMPLE_RATE
from .signal import ConstantSignal, Signal

__all__ = ('Envelope', 'ADSR', 'Decay', 'Line', 'envelope')

class Envelope(Signal):
    """
    An envelope. This signal should not be a periodic waveform, but rather a broad shape.

    You can use this base class to specify a square pulse for the given duration, in seconds.
    """
    def __init__(self, duration):
        self.duration = duration * SAMPLE_RATE
        self.pure = True

    def amplitude(self, frame: int) -> float:
        if frame < 0:
            return 0.
        if frame < self.duration:
            return 1.
        return 0.

    def apply_adsr(self, attack, decay, release, attack_level=1, sustain_level=0.5):
        """
        Return the current envelope with an additional ADSR component.
        Parameters are the same as for the ADSR class initializer.
        """
        sub = 1 if type(self) is Envelope else self
        return sub * ADSR(attack, decay, self.duration / SAMPLE_RATE - attack - decay - release, release, attack_level, sustain_level)

    def apply_decay(self, decay_param):
        """
        Add an exponential decay to the current envelope.

        :param decay_param:     The speed of the decay, between 0 and 1. Lower values decay faster.
        """
        decay = Decay(0, self.duration / SAMPLE_RATE, 0, decay_param)
        if type(self) is Envelope:
            return decay
        else:
            result = self * decay
            assert not isinstance(result, ConstantSignal)
            return result


class ADSR(Envelope):
    """
    An envelope with portions for Attack[1], Decay[2], Sustain[3], and Release[4].
    You can also specify the attack level[5] and the sustain level[6].
    Timing parameters are in seconds, level parameters are between 0 and 1.
    ::

           /\\                    [5]
          /  \\______________     [6]
         /                  \\
        /                    \\
         [1] [2]   [3]     [4]

    """
    def __init__(self, attack, decay, sustain, release, attack_level: Union[int, float]=1, sustain_level=0.5):
        super(ADSR, self).__init__(attack + decay + sustain + release)
        self.attack = attack * SAMPLE_RATE
        self.decay = decay * SAMPLE_RATE
        self.sustain = sustain * SAMPLE_RATE
        self.release = release * SAMPLE_RATE
        self.attack_level = float(attack_level)
        self.sustain_level = float(sustain_level)

    def amplitude(self, frame):
        if frame < 0:
            return 0.
        if frame < self.attack:
            return float(frame) / self.attack * self.attack_level
        frame -= self.attack
        if frame < self.decay:
            return float(frame) / self.decay * -(self.attack_level - self.sustain_level) + self.attack_level
        frame -= self.decay
        if frame < self.sustain:
            return self.sustain_level
        frame -= self.sustain
        if frame < self.release:
            return float(frame) / self.release * -self.sustain_level + self.sustain_level
        return 0.

class Decay(Envelope):
    """
    An enevelope with an attack[1], an exponential decay[2], and a release[3].
    You can specify the decay parameter[4] (between 0 and 1, smaller decays faster,
    and the height of the attack peak[5].
    Times are given in seconds.
    ::

           /*.                    [5]
          /   *._
         /       **-..___       f = [4]^x
        /                \\
         [1]       [2]   [3]
    """
    def __init__(self, attack, sustain, release, decay_param=0.5, attack_level=1.0):
        super(Decay, self).__init__(attack + sustain + release)
        self.attack = attack * SAMPLE_RATE
        self.sustain = sustain * SAMPLE_RATE
        self.release = release * SAMPLE_RATE
        self.decay_param = float(decay_param)
        self.attack_level = float(attack_level)
        self.release_level = self.decay_param ** (float(self.attack + self.sustain) / SAMPLE_RATE) * self.attack_level

    def amplitude(self, frame):
        if frame < 0:
            return 0
        if frame < self.attack:
            return float(frame) / self.attack * self.attack_level
        frame -= self.attack
        if frame < self.sustain:
            return self.decay_param ** (float(frame) / SAMPLE_RATE) * self.attack_level
        frame -= self.sustain
        if frame < self.release:
            return float(frame) / self.release * -self.release_level + self.release_level
        return 0

def envelope(sustain=None,
             attack=0.01,
             decay=None,
             release=None,
             attack_level: Union[int, float]=1,
             sustain_level=0.7,
             decaying_sustain=True):
    """
    A convenience method to produce an envelope for a musical instrument note.
    All parameters are optional and will pick sane defaults if not provided.

    :param sustain:         The length of the note (seconds)
    :param atack:           The length of the note's attack (seconds)
    :param decay:           if decaying_sustain:
                                the exponential decay parameter.
                            else:
                                the length of the note's decay (seconds)
    :param attack_level:    The intensity of the initial note attack
    :param sustain_level:   Only if not decaying_sustain, the intensity of the note's sustained body
    :param decaying_sustain: Whether the note's sustaining period should involve exponential decay
    """
    if decaying_sustain:
        if decay is None: decay = 0.5
        if release is None: release = 0.1
        if sustain is None: sustain = math.log(0.001, decay)
        return Decay(attack, sustain, release, decay, attack_level=attack_level)
    else:
        if decay is None: decay = 0.1
        if release is None: release = 0.3
        if sustain is None: sustain = 1
        return ADSR(attack, decay, sustain, release, attack_level, sustain_level)

class Line(Envelope):
    """
    A linear line running from start[1] to end[2] in a given time[3] in seconds.
    ::

        *--__                   [1]
             *--__
                  *--__
                       *--__    [2]
               [3]
    """
    def __init__(self, start, end, duration):
        super(Line, self).__init__(duration)
        self.start = start
        self.end = end

    def amplitude(self, frame):
        #if frame == 0:
            #print 'line starting!', self.start, self.end, self.duration/SAMPLE_RATE
        #if self.duration == 5 * SAMPLE_RATE:
        #    print 'AAAA', frame
        #    import ipdb; ipdb.set_trace()
        if frame < 0:
            return self.start
        if frame >= self.duration:
            return self.end
        return (float(frame) / self.duration) * (self.end - self.start) + self.start




import bisect

from .signal import Signal
from .notes import notename

class AsyncPlayer(Signal):
    """
    An object useful for asynchronous play.

    The idea is that you call `.play_async()` on it, then you can call `.play()` and `.mute()`
    on different sound objects to play or stop them in real time.
    """
    def __init__(self):
        self.frame = 0
        self.pure = False
        self.duration = float('inf')
        self.playing = []
        self.infinite = []
        self.callbacks = []
        self.to_play = []

    def amplitude(self, frame):
        for n in self.to_play:
            self._play(n)
        self.to_play = []

        out = 0.
        self.frame = frame
        while len(self.callbacks) != 0 and frame >= -self.callbacks[-1][0]:
            _, callback = self.callbacks.pop()
            if callable(callback):
                callback()
            elif type(callback) is tuple:
                if len(callback) == 2 and callable(callback[0]):
                    callback[0](*callback[1])
                else:
                    self.play(*callback)  # type: ignore
            else:
                self.play(callback)

        while len(self.playing) != 0 and frame >= -self.playing[-1][0]:
            self.playing.pop()

        for _, start, note in self.playing:
            out += note.amplitude(frame - start)
        for start, note in self.infinite:
            out += note.amplitude(frame - start)

        return out

    def _play(self, note):
        if note.duration == float('inf'):
            self.infinite.append((self.frame, note))
        else:
            end = self.frame + int(note.duration)
            bisect.insort(self.playing, (-end, self.frame, note))

    def play(self, note):  # type: ignore[reportIncompatibleMethodOverride]
        """
        :param note:    A Signal object to play, starting now.
        """
        self.to_play.append(note)

    def mute(self, note):
        """
        :param note:    The sound object to mute. Works based on object identity.
        """
        i = 0
        while i < len(self.playing):
            if self.playing[i][2] == note:
                self.playing.pop(i)
            else:
                i += 1
        i = 0
        while i < len(self.infinite):
            if self.infinite[i][1] == note:
                self.infinite.pop(i)
            else:
                i += 1

    def queue(self, when, func, relative=True):
        """
        Queue an event some number of frames in the future.

        :param when:        The number of frames in the future to perform the event
        :param func:        The thing to do in the future
        :param relative:    Optional. If set to false, `when` serves as an absolute timestamp since
                            play started instead of a relative count.

        If func is a callable object, it will be called.
        If func is a tuple, `self.play()` will be called with the tuple contents as args
        Other wise, `self.play()` will be called with func as an argument.
        """
        if relative: when += self.frame
        bisect.insort(self.callbacks, (-when, func))

class KeyedAsyncPlayer(AsyncPlayer):
    """
    An adaptation of AsyncPlayer to associate sounds with keys.
    This way, when you play another note under the same key, the first note that was played
    will be muted.
    """
    def __init__(self):
        super(KeyedAsyncPlayer, self).__init__()
        self.active = {}

    def play(self, note, name=None):    # pylint: disable=arguments-differ
        """
        :param note:    The sound to play
        :param name:    The key to play the sound under
        """
        self.mute(name)
        self.active[name] = note
        super(KeyedAsyncPlayer, self).play(note)

    def mute(self, note):
        """
        :param note:    The key of the sound to mute
        """
        if note not in self.active:
            return
        super(KeyedAsyncPlayer, self).mute(self.active.pop(note))

class InstrumentPlayer(KeyedAsyncPlayer):
    """
    An extension of KeyedAsyncPlayer where the keys are the frequencies of the notes being played,
    and `.play()` accepts arguments to an `Instrument.note()` call.

    The idea is that this simulates an instrument where when you play two of the same note,
    the second note reuses the physical resource producing the sound for that note.
    """
    def __init__(self, instrument):
        """
        :param instrument:  The instrument to play with
        """
        super(InstrumentPlayer, self).__init__()
        self.instrument = instrument

    def play(self, note, *args, **kwargs):
        """
        All parameters are passed directly through to `instrument.note()`
        """
        sig = self.instrument(note, *args, **kwargs)
        super(InstrumentPlayer, self).play(note, sig)

class GuitarStrummer(KeyedAsyncPlayer):
    """
    A guitar that you can strum in various chords!
    """
    def __init__(self, sample):
        super(GuitarStrummer, self).__init__()
        self.sample = sample

    chords = {
        'C': [0, 3, 2, 0, 1, 0],
        'Cm': [0, 0, 5, 5, 4, 3],
        'D': [None, 0, 0, 2, 3, 2],
        'Dm': [0, 0, 0, 2, 3, 2],
        'E': [0, 2, 2, 1, 0, 0],
        'Em': [0, 2, 2, 0, 0, 0],
        'F': [1, 3, 3, 2, 1, 1],
        'Fm': [0, 0, 3, 1, 1, 1],
        'G': [3, 2, 0, 0, 0, 3],
        'Gm': [0, 0, 5, 3, 3, 3],
        'A': [0, 0, 2, 2, 2, 0],
        'Am': [0, 0, 2, 2, 1, 0],
        'A7': [0, 0, 2, 0, 2, 0],
        'Am7': [0, 0, 2, 0, 1, 0],
        'B': [0, 0, 4, 4, 4, 2]
    }

    base_frequencies = list(map(notename, ['E2', 'A2', 'D3', 'G3', 'B3', 'E4']))

    def strum_down(self, chord, delay=200):
        """
        Strum the guitar in the given chord from top to bottom

        :param chord:   The chord to play, as a string. Look at the source for this class
                        for a list of supported chords.
        :param delay:   The time inbetween chord plucks, in frames. Optional.
        """
        freqs = [None if n is None else 2**(n/12.)*b for n, b in zip(self.chords[chord], self.base_frequencies)]
        delays = [n * delay for n in range(6)]
        for i, (f, d) in enumerate(zip(freqs, delays)):
            if f is not None:
                self.queue(d, (self.sample(f), i))

    def strum_up(self, chord, delay=200):
        """
        Strum the guitar in the given chord from bottom to top

        :param chord:   The chord to play, as a string. Look at the source for this class
                        for a list of supported chords.
        :param delay:   The time inbetween chord plucks, in frames. Optional.
        """
        freqs = [None if n is None else 2**(n/12.)*b for n, b in zip(self.chords[chord], self.base_frequencies)]
        delays = [n * delay for n in reversed(range(6))]
        for i, (f, d) in enumerate(zip(freqs, delays)):
            if f is not None:
                self.queue(d, (self.sample(f), i))
            else:
                self.queue(d, (self.mute, (i,)))
